@@ -25,6 +25,7 @@ use crate::{
     ast::{self, Visitor},
     ast_map::{AstMap, AstNode},
     common::{arenas::Alloc, arenas::TypedArena, Session},
+    func_args::FuncArgList,
     hir::{self, HirNode},
     port_list::PortList,
     resolver::Scope,
@@ -50,6 +51,8 @@ pub struct GlobalContext<'gcx> {
     ast_map: AstMap<'gcx>,
     /// The AST nodes.
     ast_map2: RefCell<HashMap<NodeId, &'gcx dyn ast::AnyNode<'gcx>>>,
+    /// The AST roots.
+    roots: RefCell<Vec<&'gcx ast::Root<'gcx>>>,
     /// The modules in the AST.
     modules: RefCell<HashMap<Name, NodeId>>,
     /// The packages in the AST.
@@ -73,6 +76,7 @@ impl<'gcx> GlobalContext<'gcx> {
             storage: Default::default(),
             ast_map: Default::default(),
             ast_map2: Default::default(),
+            roots: Default::default(),
             modules: Default::default(),
             packages: Default::default(),
             interfaces: Default::default(),
@@ -87,6 +91,11 @@ impl<'gcx> GlobalContext<'gcx> {
     /// Use the `find_global_item` function afterwards to look up the id of
     /// modules that were added.
     pub fn add_root(&self, root: &'gcx ast::Root<'gcx>) {
+        // Add the root node to the list of roots registered with this context,
+        // such that we can find them again afterwards.
+        self.roots.borrow_mut().push(root);
+
+        // Link up the AST. This establishes the parent back-arcs in the AST.
         debug!("Linking nodes");
         let mut index = 0;
         root.link(None, &mut index);
@@ -98,9 +107,11 @@ impl<'gcx> GlobalContext<'gcx> {
 
         // Register nodes with the AST map. This is a necessary hack until
         // we have moved away from querying nodes merely by ID.
+        // TODO(fschuiki): Remove this once we no longer look for AST nodes
+        // based on their ID.
         self.register_ast(root);
 
-        // Resolve names for debugging purposes.
+        // Check for name conflicts.
         debug!("Checking names");
         self.nameck(root);
 
@@ -147,6 +158,11 @@ impl<'gcx> GlobalContext<'gcx> {
         self.add_root(root);
     }
 
+    /// Get an iterator over all roots in the AST.
+    pub fn roots(&self) -> impl Iterator<Item = &'gcx ast::Root<'gcx>> {
+        self.roots.borrow().clone().into_iter()
+    }
+
     /// Find a module in the AST.
     pub fn find_module(&self, name: Name) -> Option<NodeId> {
         self.modules.borrow().get(&name).cloned()
@@ -269,6 +285,7 @@ pub struct GlobalArenas<'t> {
     ast_exprs: TypedArena<ast::Expr<'t>>,
     /// The underlying storage for type operations.
     type_storage: ty::TypeStorage<'t>,
+    func_arg_lists: TypedArena<FuncArgList<'t>>,
 }
 
 impl<'t> GlobalArenas<'t> {
@@ -330,6 +347,11 @@ impl<'t> GlobalArenas<'t> {
     pub fn alloc_ast_expr(&'t self, ast: ast::Expr<'t>) -> &'t ast::Expr {
         self.ast_exprs.alloc(ast)
     }
+
+    /// Allocate a func arg list.
+    pub fn alloc_func_arg_list(&'t self, func_arg_list: FuncArgList<'t>) -> &'t FuncArgList<'t> {
+        self.func_arg_lists.alloc(func_arg_list)
+    }
 }
 
 /// Allow AST nodes to be allocated into `GlobalArenas`.
@@ -471,19 +493,25 @@ pub trait Context<'gcx>: DiagEmitter + QueryDatabase<'gcx> + ty::HasTypeStorage<
             Some(node) => return Ok(node),
             None => (),
         }
-        if let Some(&span) = self.gcx().node_id_to_span.borrow().get(&node_id) {
-            self.emit(
-                DiagBuilder2::bug(format!("no ast node for `{}` in the map", span.extract()))
-                    .span(span),
-            );
-            error!("Offending node: {:?}", node_id);
-        } else {
-            self.emit(DiagBuilder2::bug(format!(
-                "no ast node for {:?} in the map",
-                node_id
-            )));
-        }
-        panic!("Other map contains: {:?}", other);
+        self.emit(
+            DiagBuilder2::bug(format!(
+                "no ast node for `{}` in the map",
+                other.span().extract()
+            ))
+            .span(other.span())
+            .add_note(format!("Offending node: {:#1?}", other))
+            .add_note(format!("Required here:\n{:?}", {
+                let v: Vec<_> = backtrace::Backtrace::new()
+                    .frames()
+                    .iter()
+                    .cloned()
+                    .skip(1)
+                    .take(8)
+                    .collect();
+                backtrace::Backtrace::from(v)
+            })),
+        );
+        panic!("no ast node");
     }
 
     /// Register an `ast::AnyNode` for later retrieval by ID.
@@ -697,6 +725,9 @@ impl<'a, 'b> ast::Visitor<'a> for AstMapRegistrator<'a, 'b> {
         self.cx.gcx().ast_map2.borrow_mut().insert(node.id(), node);
 
         for n in AstNode::from_all(node.as_all()) {
+            if let AstNode::Any(_) = n {
+                continue;
+            }
             let parent = n.get_any().unwrap().get_parent().unwrap();
             let parent = if let Some(kind) = parent.as_all().get_type_kind() {
                 kind.get_parent().unwrap()
