@@ -3,7 +3,7 @@
 //! Lowering of AST nodes to HIR nodes.
 
 use crate::crate_prelude::*;
-use crate::{ast_map::AstNode, hir::HirNode};
+use crate::{ast_map::AstNode, hir::HirNode, resolver::DefNode};
 use bit_vec::BitVec;
 use num::BigInt;
 
@@ -187,190 +187,7 @@ pub(crate) fn hir_of<'a>(cx: &impl Context<'a>, node_id: NodeId) -> Result<HirNo
             };
             Ok(HirNode::Proc(cx.arena().alloc_hir(hir)))
         }
-        AstNode::Stmt(stmt) => {
-            let kind = match stmt.kind {
-                ast::NullStmt => hir::StmtKind::Null,
-                ast::SequentialBlock(ref stmts) => {
-                    let mut next_rib = node_id;
-                    hir::StmtKind::Block(
-                        stmts
-                            .iter()
-                            .map(|stmt| {
-                                let id = cx.map_ast_with_parent(AstNode::Stmt(stmt), next_rib);
-                                next_rib = id;
-                                id
-                            })
-                            .collect(),
-                    )
-                }
-                ast::BlockingAssignStmt {
-                    ref lhs,
-                    ref rhs,
-                    op,
-                } => hir::StmtKind::Assign {
-                    lhs: cx.map_ast_with_parent(AstNode::Expr(lhs), node_id),
-                    rhs: cx.map_ast_with_parent(AstNode::Expr(rhs), node_id),
-                    kind: hir::AssignKind::Block(op),
-                },
-                ast::TimedStmt(ref control, ref inner_stmt) => {
-                    let control = match *control {
-                        ast::TimingControl::Delay(ref dc) => hir::TimingControl::Delay(
-                            cx.map_ast_with_parent(AstNode::Expr(&dc.expr), node_id),
-                        ),
-                        ast::TimingControl::Event(ref ec) => match ec.data {
-                            ast::EventControlData::Implicit => hir::TimingControl::ImplicitEvent,
-                            ast::EventControlData::Expr(ref expr) => {
-                                hir::TimingControl::ExplicitEvent(
-                                    cx.map_ast_with_parent(AstNode::EventExpr(expr), node_id),
-                                )
-                            }
-                        },
-                        _ => {
-                            debug!("{:#?}", stmt);
-                            bug_span!(
-                                stmt.span(),
-                                cx,
-                                "lowering of timing control {} to hir not implemented",
-                                stmt
-                            );
-                        }
-                    };
-                    hir::StmtKind::Timed {
-                        control,
-                        stmt: cx.map_ast_with_parent(AstNode::Stmt(inner_stmt), node_id),
-                    }
-                }
-                ast::IfStmt {
-                    ref cond,
-                    ref main_stmt,
-                    ref else_stmt,
-                    ..
-                } => hir::StmtKind::If {
-                    cond: cx.map_ast_with_parent(AstNode::Expr(cond), node_id),
-                    main_stmt: cx.map_ast_with_parent(AstNode::Stmt(main_stmt), node_id),
-                    else_stmt: else_stmt
-                        .as_ref()
-                        .map(|else_stmt| cx.map_ast_with_parent(AstNode::Stmt(else_stmt), node_id)),
-                },
-                ast::ExprStmt(ref expr) => {
-                    hir::StmtKind::Expr(cx.map_ast_with_parent(AstNode::Expr(expr), node_id))
-                }
-                ast::ForeverStmt(ref body) => hir::StmtKind::Loop {
-                    kind: hir::LoopKind::Forever,
-                    body: cx.map_ast_with_parent(AstNode::Stmt(body), node_id),
-                },
-                ast::RepeatStmt(ref count, ref body) => hir::StmtKind::Loop {
-                    kind: hir::LoopKind::Repeat(
-                        cx.map_ast_with_parent(AstNode::Expr(count), node_id),
-                    ),
-                    body: cx.map_ast_with_parent(AstNode::Stmt(body), node_id),
-                },
-                ast::WhileStmt(ref cond, ref body) => hir::StmtKind::Loop {
-                    kind: hir::LoopKind::While(
-                        cx.map_ast_with_parent(AstNode::Expr(cond), node_id),
-                    ),
-                    body: cx.map_ast_with_parent(AstNode::Stmt(body), node_id),
-                },
-                ast::DoStmt(ref body, ref cond) => hir::StmtKind::Loop {
-                    kind: hir::LoopKind::Do(cx.map_ast_with_parent(AstNode::Expr(cond), node_id)),
-                    body: cx.map_ast_with_parent(AstNode::Stmt(body), node_id),
-                },
-                ast::ForStmt(ref init, ref cond, ref step, ref body) => {
-                    let init = cx.map_ast_with_parent(AstNode::Stmt(init), node_id);
-                    let cond = cx.map_ast_with_parent(AstNode::Expr(cond), init);
-                    let step = cx.map_ast_with_parent(AstNode::Expr(step), init);
-                    hir::StmtKind::Loop {
-                        kind: hir::LoopKind::For(init, cond, step),
-                        body: cx.map_ast_with_parent(AstNode::Stmt(body), init),
-                    }
-                }
-                ast::VarDeclStmt(ref decls) => {
-                    let mut stmts = vec![];
-                    let parent = cx.parent_node_id(node_id).unwrap();
-                    let rib = alloc_var_decl(cx, decls, parent, &mut stmts);
-                    hir::StmtKind::InlineGroup { stmts, rib }
-                }
-                ast::NonblockingAssignStmt {
-                    ref lhs,
-                    ref rhs,
-                    ref delay,
-                    ..
-                } => hir::StmtKind::Assign {
-                    lhs: cx.map_ast_with_parent(AstNode::Expr(lhs), node_id),
-                    rhs: cx.map_ast_with_parent(AstNode::Expr(rhs), node_id),
-                    kind: match *delay {
-                        Some(ref dc) => hir::AssignKind::NonblockDelay(
-                            cx.map_ast_with_parent(AstNode::Expr(&dc.expr), node_id),
-                        ),
-                        None => hir::AssignKind::Nonblock,
-                    },
-                },
-                ast::CaseStmt {
-                    ref expr,
-                    mode: ast::CaseMode::Normal,
-                    ref items,
-                    kind,
-                    ..
-                } => {
-                    let expr = cx.map_ast_with_parent(AstNode::Expr(expr), node_id);
-                    let mut ways = vec![];
-                    let mut default = None;
-                    for item in items {
-                        match *item {
-                            ast::CaseItem::Default(ref stmt) => {
-                                if default.is_none() {
-                                    default =
-                                        Some(cx.map_ast_with_parent(AstNode::Stmt(stmt), node_id));
-                                } else {
-                                    cx.emit(
-                                        DiagBuilder2::error("multiple default cases")
-                                            .span(stmt.human_span()),
-                                    );
-                                }
-                            }
-                            ast::CaseItem::Expr(ref exprs, ref stmt) => ways.push((
-                                exprs
-                                    .iter()
-                                    .map(|expr| {
-                                        cx.map_ast_with_parent(AstNode::Expr(expr), node_id)
-                                    })
-                                    .collect(),
-                                cx.map_ast_with_parent(AstNode::Stmt(stmt), node_id),
-                            )),
-                        }
-                    }
-                    hir::StmtKind::Case {
-                        expr,
-                        ways,
-                        default,
-                        kind,
-                    }
-                }
-                ast::AssertionStmt { .. } => {
-                    cx.emit(
-                        DiagBuilder2::warning("unsupported: immediate assertion; ignored")
-                            .span(stmt.human_span()),
-                    );
-                    hir::StmtKind::Null
-                }
-                _ => {
-                    error!("{:#?}", stmt);
-                    bug_span!(
-                        stmt.span(),
-                        cx,
-                        "lowering of {:?} to hir not implemented",
-                        stmt
-                    );
-                }
-            };
-            let hir = hir::Stmt {
-                id: node_id,
-                label: stmt.label.map(|n| Spanned::new(n, stmt.span)), // this is horrible...
-                span: stmt.span,
-                kind: kind,
-            };
-            Ok(HirNode::Stmt(cx.arena().alloc_hir(hir)))
-        }
+        AstNode::Stmt(stmt) => cx.hir_of_stmt(Ref(stmt)).map(HirNode::Stmt),
         AstNode::EventExpr(expr) => {
             let mut events = vec![];
             lower_event_expr(cx, expr, node_id, &mut events, &mut vec![])?;
@@ -486,23 +303,22 @@ pub(crate) fn hir_of<'a>(cx: &impl Context<'a>, node_id: NodeId) -> Result<HirNo
             Ok(HirNode::EnumVariant(cx.arena().alloc_hir(hir)))
         }
         AstNode::Import(import) => unreachable!("import should never be lowered: {:#?}", import),
-        AstNode::SubroutineDecl(decl) => {
-            let hir = hir::Subroutine {
-                id: node_id,
-                name: decl.prototype.name,
-                span: decl.span,
-                kind: decl.prototype.kind,
-                retty: decl
-                    .prototype
-                    .retty
-                    .as_ref()
-                    .map(|ty| cx.map_ast_with_parent(AstNode::Type(ty), node_id)),
-            };
-            Ok(HirNode::Subroutine(cx.arena().alloc_hir(hir)))
-        }
+        AstNode::CallArg(x) => Ok(HirNode::CallArg(x)),
+        AstNode::Any(ast) => match ast.as_all() {
+            ast::AllNode::SubroutinePort(x) => Ok(HirNode::SubroutinePort(x)),
+            _ => {
+                error!("{:#?}", ast);
+                bug_span!(ast.span(), cx, "lowering of {} to hir not implemented", ast);
+            }
+        },
         _ => {
             error!("{:#?}", ast);
-            cx.unimp_msg("lowering of", &ast)
+            bug_span!(
+                ast.span(),
+                cx,
+                "lowering of {} to hir not implemented",
+                ast.desc_full()
+            );
         }
     }
 }
@@ -1159,98 +975,7 @@ fn lower_expr_inner<'gcx>(
             // };
             hir::ExprKind::Index(indexee, mode)
         }
-        ast::CallExpr(ref callee, ref args) => match callee.data {
-            ast::SysIdentExpr(ident) => {
-                let map_unary = || match args.as_slice() {
-                    [ast::CallArg {
-                        expr: Some(ref arg),
-                        ..
-                    }] => Ok(arg),
-                    _ => {
-                        cx.emit(
-                            DiagBuilder2::error(format!("`{}` takes one argument", ident))
-                                .span(expr.human_span()),
-                        );
-                        Err(())
-                    }
-                };
-                let map_unary_id =
-                    || Ok(cx.map_ast_with_parent(AstNode::Expr(map_unary()?), node_id));
-                let map_array_dim = |func| match args.as_slice() {
-                    [ast::CallArg {
-                        expr: Some(ref arg),
-                        ..
-                    }] => Ok(hir::BuiltinCall::ArrayDim(func, arg, None)),
-                    [ast::CallArg {
-                        expr: Some(ref arg),
-                        ..
-                    }, ast::CallArg {
-                        expr: Some(ref dim),
-                        ..
-                    }] => Ok(hir::BuiltinCall::ArrayDim(func, arg, Some(dim))),
-                    _ => {
-                        cx.emit(
-                            DiagBuilder2::error(format!("`{}` takes one or two arguments", ident))
-                                .span(expr.human_span()),
-                        );
-                        Err(())
-                    }
-                };
-                hir::ExprKind::Builtin(match &*ident.value.as_str() {
-                    "clog2" => hir::BuiltinCall::Clog2(map_unary_id()?),
-                    "signed" => hir::BuiltinCall::Signed(map_unary_id()?),
-                    "unsigned" => hir::BuiltinCall::Unsigned(map_unary_id()?),
-                    "countones" => hir::BuiltinCall::CountOnes(map_unary()?),
-                    "onehot" => hir::BuiltinCall::OneHot(map_unary()?),
-                    "onehot0" => hir::BuiltinCall::OneHot0(map_unary()?),
-                    "isunknown" => hir::BuiltinCall::IsUnknown(map_unary()?),
-                    "left" => map_array_dim(hir::ArrayDim::Left)?,
-                    "right" => map_array_dim(hir::ArrayDim::Right)?,
-                    "low" => map_array_dim(hir::ArrayDim::Low)?,
-                    "high" => map_array_dim(hir::ArrayDim::High)?,
-                    "increment" => map_array_dim(hir::ArrayDim::Increment)?,
-                    "size" => map_array_dim(hir::ArrayDim::Size)?,
-                    "display" | "info" | "warning" | "error" | "fatal" => {
-                        cx.emit(
-                            DiagBuilder2::warning(format!(
-                                "unsupported: system task `${}`; ignored",
-                                ident
-                            ))
-                            .span(expr.human_span()),
-                        );
-                        hir::BuiltinCall::Unsupported
-                    }
-                    _ => {
-                        cx.emit(
-                            DiagBuilder2::error(format!("unknown system task `${}`", ident))
-                                .span(expr.human_span()),
-                        );
-                        hir::BuiltinCall::Unsupported
-                    }
-                })
-            }
-            ast::IdentExpr(name) => {
-                let target =
-                    cx.resolve_upwards_or_error(name, cx.parent_node_id(node_id).unwrap())?;
-                hir::ExprKind::FunctionCall(
-                    target,
-                    args.iter()
-                        .map(|arg| lower_call_arg(cx, arg, node_id))
-                        .collect(),
-                )
-            }
-            _ => {
-                error!("{:#?}", callee);
-                cx.emit(
-                    DiagBuilder2::warning(format!(
-                        "`{}` is not something that can be called",
-                        expr.span().extract()
-                    ))
-                    .span(expr.human_span()),
-                );
-                return Err(());
-            }
-        },
+        ast::CallExpr(ref callee, ref args) => lower_call(cx, expr, callee, args)?,
         ast::TernaryExpr {
             ref cond,
             ref true_expr,
@@ -1445,15 +1170,188 @@ fn lower_expr_inner<'gcx>(
             lhs: lhs.as_ref(),
             rhs: rhs.as_ref(),
         },
-        _ => {
-            error!("{:#1?}", expr);
-            bug_span!(
-                expr.span,
-                cx,
-                "lowering of {:?} to hir not implemented",
-                expr
-            );
+        _ => hir::ExprKind::Ast(expr),
+    })
+}
+
+/// Lower an AST statement to HIR.
+#[moore_derive::query]
+pub(crate) fn hir_of_stmt<'a>(
+    cx: &impl Context<'a>,
+    Ref(ast): Ref<'a, ast::Stmt<'a>>,
+) -> Result<&'a hir::Stmt<'a>> {
+    let kind = lower_stmt_inner(cx, ast)?;
+    let hir = hir::Stmt {
+        id: ast.id(),
+        label: ast.label.map(|n| Spanned::new(n, ast.span)), // this is horrible...
+        span: ast.span,
+        kind: kind,
+    };
+    Ok(cx.arena().alloc_hir(hir))
+}
+
+fn lower_stmt_inner<'a>(
+    cx: &impl Context<'a>,
+    stmt: &'a ast::Stmt<'a>,
+) -> Result<hir::StmtKind<'a>> {
+    let node_id = stmt.id();
+    Ok(match stmt.kind {
+        ast::NullStmt => hir::StmtKind::Null,
+        ast::SequentialBlock(ref stmts) => {
+            let mut next_rib = node_id;
+            hir::StmtKind::Block(
+                stmts
+                    .iter()
+                    .map(|stmt| {
+                        let id = cx.map_ast_with_parent(AstNode::Stmt(stmt), next_rib);
+                        next_rib = id;
+                        id
+                    })
+                    .collect(),
+            )
         }
+        ast::BlockingAssignStmt {
+            ref lhs,
+            ref rhs,
+            op,
+        } => hir::StmtKind::Assign {
+            lhs: cx.map_ast_with_parent(AstNode::Expr(lhs), node_id),
+            rhs: cx.map_ast_with_parent(AstNode::Expr(rhs), node_id),
+            kind: hir::AssignKind::Block(op),
+        },
+        ast::TimedStmt(ref control, ref inner_stmt) => {
+            let control = match *control {
+                ast::TimingControl::Delay(ref dc) => hir::TimingControl::Delay(
+                    cx.map_ast_with_parent(AstNode::Expr(&dc.expr), node_id),
+                ),
+                ast::TimingControl::Event(ref ec) => match ec.data {
+                    ast::EventControlData::Implicit => hir::TimingControl::ImplicitEvent,
+                    ast::EventControlData::Expr(ref expr) => hir::TimingControl::ExplicitEvent(
+                        cx.map_ast_with_parent(AstNode::EventExpr(expr), node_id),
+                    ),
+                },
+                _ => {
+                    debug!("{:#?}", stmt);
+                    bug_span!(
+                        stmt.span(),
+                        cx,
+                        "lowering of timing control {} to hir not implemented",
+                        stmt
+                    );
+                }
+            };
+            hir::StmtKind::Timed {
+                control,
+                stmt: cx.map_ast_with_parent(AstNode::Stmt(inner_stmt), node_id),
+            }
+        }
+        ast::IfStmt {
+            ref cond,
+            ref main_stmt,
+            ref else_stmt,
+            ..
+        } => hir::StmtKind::If {
+            cond: cx.map_ast_with_parent(AstNode::Expr(cond), node_id),
+            main_stmt: cx.map_ast_with_parent(AstNode::Stmt(main_stmt), node_id),
+            else_stmt: else_stmt
+                .as_ref()
+                .map(|else_stmt| cx.map_ast_with_parent(AstNode::Stmt(else_stmt), node_id)),
+        },
+        ast::ExprStmt(ref expr) => {
+            hir::StmtKind::Expr(cx.map_ast_with_parent(AstNode::Expr(expr), node_id))
+        }
+        ast::ForeverStmt(ref body) => hir::StmtKind::Loop {
+            kind: hir::LoopKind::Forever,
+            body: cx.map_ast_with_parent(AstNode::Stmt(body), node_id),
+        },
+        ast::RepeatStmt(ref count, ref body) => hir::StmtKind::Loop {
+            kind: hir::LoopKind::Repeat(cx.map_ast_with_parent(AstNode::Expr(count), node_id)),
+            body: cx.map_ast_with_parent(AstNode::Stmt(body), node_id),
+        },
+        ast::WhileStmt(ref cond, ref body) => hir::StmtKind::Loop {
+            kind: hir::LoopKind::While(cx.map_ast_with_parent(AstNode::Expr(cond), node_id)),
+            body: cx.map_ast_with_parent(AstNode::Stmt(body), node_id),
+        },
+        ast::DoStmt(ref body, ref cond) => hir::StmtKind::Loop {
+            kind: hir::LoopKind::Do(cx.map_ast_with_parent(AstNode::Expr(cond), node_id)),
+            body: cx.map_ast_with_parent(AstNode::Stmt(body), node_id),
+        },
+        ast::ForStmt(ref init, ref cond, ref step, ref body) => {
+            let init = cx.map_ast_with_parent(AstNode::Stmt(init), node_id);
+            let cond = cx.map_ast_with_parent(AstNode::Expr(cond), init);
+            let step = cx.map_ast_with_parent(AstNode::Expr(step), init);
+            hir::StmtKind::Loop {
+                kind: hir::LoopKind::For(init, cond, step),
+                body: cx.map_ast_with_parent(AstNode::Stmt(body), init),
+            }
+        }
+        ast::VarDeclStmt(ref decls) => {
+            let mut stmts = vec![];
+            let parent = cx.parent_node_id(node_id).unwrap();
+            let rib = alloc_var_decl(cx, decls, parent, &mut stmts);
+            hir::StmtKind::InlineGroup { stmts, rib }
+        }
+        ast::NonblockingAssignStmt {
+            ref lhs,
+            ref rhs,
+            ref delay,
+            ..
+        } => hir::StmtKind::Assign {
+            lhs: cx.map_ast_with_parent(AstNode::Expr(lhs), node_id),
+            rhs: cx.map_ast_with_parent(AstNode::Expr(rhs), node_id),
+            kind: match *delay {
+                Some(ref dc) => hir::AssignKind::NonblockDelay(
+                    cx.map_ast_with_parent(AstNode::Expr(&dc.expr), node_id),
+                ),
+                None => hir::AssignKind::Nonblock,
+            },
+        },
+        ast::CaseStmt {
+            ref expr,
+            mode: ast::CaseMode::Normal,
+            ref items,
+            kind,
+            ..
+        } => {
+            let expr = cx.map_ast_with_parent(AstNode::Expr(expr), node_id);
+            let mut ways = vec![];
+            let mut default = None;
+            for item in items {
+                match *item {
+                    ast::CaseItem::Default(ref stmt) => {
+                        if default.is_none() {
+                            default = Some(cx.map_ast_with_parent(AstNode::Stmt(stmt), node_id));
+                        } else {
+                            cx.emit(
+                                DiagBuilder2::error("multiple default cases")
+                                    .span(stmt.human_span()),
+                            );
+                        }
+                    }
+                    ast::CaseItem::Expr(ref exprs, ref stmt) => ways.push((
+                        exprs
+                            .iter()
+                            .map(|expr| cx.map_ast_with_parent(AstNode::Expr(expr), node_id))
+                            .collect(),
+                        cx.map_ast_with_parent(AstNode::Stmt(stmt), node_id),
+                    )),
+                }
+            }
+            hir::StmtKind::Case {
+                expr,
+                ways,
+                default,
+                kind,
+            }
+        }
+        ast::AssertionStmt { .. } => {
+            cx.emit(
+                DiagBuilder2::warning("unsupported: immediate assertion; ignored")
+                    .span(stmt.human_span()),
+            );
+            hir::StmtKind::Null
+        }
+        _ => hir::StmtKind::Ast(stmt),
     })
 }
 
@@ -1711,18 +1609,141 @@ pub(crate) fn lower_index_mode<'gcx>(
     }
 }
 
-/// Lower a function or method call argument to HIR.
-fn lower_call_arg<'gcx>(
-    cx: &impl Context<'gcx>,
-    ast: &'gcx ast::CallArg<'gcx>,
-    parent: NodeId,
-) -> hir::CallArg {
-    hir::CallArg {
-        span: ast.span,
-        name: ast.name.map(|n| Spanned::new(n, ast.name_span)),
-        expr: ast
-            .expr
-            .as_ref()
-            .map(|expr| cx.map_ast_with_parent(AstNode::Expr(expr), parent)),
-    }
+/// Lower a function or method call.
+fn lower_call<'a>(
+    cx: &impl Context<'a>,
+    expr: &'a ast::Expr<'a>,
+    callee: &'a ast::Expr<'a>,
+    args: &'a [ast::CallArg<'a>],
+) -> Result<hir::ExprKind<'a>> {
+    Ok(match callee.data {
+        ast::SysIdentExpr(ident) => {
+            let map_unary = || match args {
+                [ast::CallArg {
+                    data:
+                        ast::CallArgData {
+                            expr: Some(ref arg),
+                            ..
+                        },
+                    ..
+                }] => Ok(arg),
+                _ => {
+                    cx.emit(
+                        DiagBuilder2::error(format!("`{}` takes one argument", ident))
+                            .span(expr.human_span()),
+                    );
+                    Err(())
+                }
+            };
+            let map_unary_id =
+                || Ok(cx.map_ast_with_parent(AstNode::Expr(map_unary()?), expr.id()));
+            let map_array_dim = |func| match args {
+                [ast::CallArg {
+                    data:
+                        ast::CallArgData {
+                            expr: Some(ref arg),
+                            ..
+                        },
+                    ..
+                }] => Ok(hir::BuiltinCall::ArrayDim(func, arg, None)),
+                [ast::CallArg {
+                    data:
+                        ast::CallArgData {
+                            expr: Some(ref arg),
+                            ..
+                        },
+                    ..
+                }, ast::CallArg {
+                    data:
+                        ast::CallArgData {
+                            expr: Some(ref dim),
+                            ..
+                        },
+                    ..
+                }] => Ok(hir::BuiltinCall::ArrayDim(func, arg, Some(dim))),
+                _ => {
+                    cx.emit(
+                        DiagBuilder2::error(format!("`{}` takes one or two arguments", ident))
+                            .span(expr.human_span()),
+                    );
+                    Err(())
+                }
+            };
+            hir::ExprKind::Builtin(match &*ident.value.as_str() {
+                "clog2" => hir::BuiltinCall::Clog2(map_unary_id()?),
+                "signed" => hir::BuiltinCall::Signed(map_unary_id()?),
+                "unsigned" => hir::BuiltinCall::Unsigned(map_unary_id()?),
+                "countones" => hir::BuiltinCall::CountOnes(map_unary()?),
+                "onehot" => hir::BuiltinCall::OneHot(map_unary()?),
+                "onehot0" => hir::BuiltinCall::OneHot0(map_unary()?),
+                "isunknown" => hir::BuiltinCall::IsUnknown(map_unary()?),
+                "left" => map_array_dim(hir::ArrayDim::Left)?,
+                "right" => map_array_dim(hir::ArrayDim::Right)?,
+                "low" => map_array_dim(hir::ArrayDim::Low)?,
+                "high" => map_array_dim(hir::ArrayDim::High)?,
+                "increment" => map_array_dim(hir::ArrayDim::Increment)?,
+                "size" => map_array_dim(hir::ArrayDim::Size)?,
+                _ => {
+                    cx.emit(
+                        DiagBuilder2::warning(format!(
+                            "unsupported: system task `${}`; ignored",
+                            ident
+                        ))
+                        .span(expr.human_span()),
+                    );
+                    hir::BuiltinCall::Unsupported
+                }
+            })
+        }
+        ast::IdentExpr(name) => {
+            // Resolve the function name and make sure it's something we can
+            // call.
+            let def = cx.resolve_local_or_error(name, cx.scope_location(expr), false)?;
+            let target = match def.node {
+                DefNode::Ast(ast) => match ast.as_all() {
+                    ast::AllNode::SubroutineDecl(x) => Some(x),
+                    _ => None,
+                },
+                DefNode::IntPort(..) => None,
+            };
+
+            // Complain if it's something we can't call.
+            let target = match target {
+                Some(x) => x,
+                None => {
+                    error!("{:#?}", target);
+                    cx.emit(
+                        DiagBuilder2::error(format!(
+                            "uncallable: {} is not a function/task",
+                            def.node
+                        ))
+                        .span(name.span())
+                        .add_note(format!("`{}` defined here:", def.name))
+                        .span(def.name.span),
+                    );
+                    return Err(());
+                }
+            };
+            debug!("Call to `{}` resolved to {}", name, target);
+
+            // TODO(fschuiki): Get rid of this at some point.
+            for arg in args {
+                cx.map_ast_with_parent(AstNode::CallArg(arg), expr.id());
+            }
+
+            // Package the call up.
+            hir::ExprKind::FunctionCall(target, args)
+        }
+        _ => {
+            error!("{:#?}", callee);
+            cx.emit(
+                DiagBuilder2::error(format!(
+                    "`{}` is not something that can be called",
+                    callee.span().extract()
+                ))
+                .span(expr.human_span()),
+            );
+            return Err(());
+        }
+    })
 }
