@@ -25,7 +25,7 @@ pub enum HirNode<'a> {
     ValueParam(&'a ValueParam),
     VarDecl(&'a VarDecl),
     Proc(&'a Proc),
-    Stmt(&'a Stmt),
+    Stmt(&'a Stmt<'a>),
     EventExpr(&'a EventExpr),
     Gen(&'a Gen),
     GenvarDecl(&'a GenvarDecl),
@@ -33,7 +33,8 @@ pub enum HirNode<'a> {
     Assign(&'a Assign),
     Package(&'a Package),
     EnumVariant(&'a EnumVariant),
-    Subroutine(&'a Subroutine),
+    SubroutinePort(&'a ast::SubroutinePort<'a>),
+    CallArg(&'a ast::CallArg<'a>),
 }
 
 impl<'hir> HasSpan for HirNode<'hir> {
@@ -59,7 +60,8 @@ impl<'hir> HasSpan for HirNode<'hir> {
             HirNode::Assign(x) => x.span(),
             HirNode::Package(x) => x.span(),
             HirNode::EnumVariant(x) => x.span(),
-            HirNode::Subroutine(x) => x.span(),
+            HirNode::SubroutinePort(x) => x.span(),
+            HirNode::CallArg(x) => x.span(),
         }
     }
 
@@ -85,7 +87,8 @@ impl<'hir> HasSpan for HirNode<'hir> {
             HirNode::Assign(x) => x.human_span(),
             HirNode::Package(x) => x.human_span(),
             HirNode::EnumVariant(x) => x.human_span(),
-            HirNode::Subroutine(x) => x.human_span(),
+            HirNode::SubroutinePort(x) => x.human_span(),
+            HirNode::CallArg(x) => x.human_span(),
         }
     }
 }
@@ -113,7 +116,8 @@ impl<'hir> HasDesc for HirNode<'hir> {
             HirNode::Assign(x) => x.desc(),
             HirNode::Package(x) => x.desc(),
             HirNode::EnumVariant(x) => x.desc(),
-            HirNode::Subroutine(x) => x.desc(),
+            HirNode::SubroutinePort(..) => "subroutine port",
+            HirNode::CallArg(..) => "call argument",
         }
     }
 
@@ -139,7 +143,8 @@ impl<'hir> HasDesc for HirNode<'hir> {
             HirNode::Assign(x) => x.desc_full(),
             HirNode::Package(x) => x.desc_full(),
             HirNode::EnumVariant(x) => x.desc_full(),
-            HirNode::Subroutine(x) => x.desc_full(),
+            HirNode::SubroutinePort(x) => x.to_string(),
+            HirNode::CallArg(x) => x.to_string(),
         }
     }
 }
@@ -674,13 +679,15 @@ pub enum ExprKind<'a> {
     /// An inside expression such as `a inside {b, c}`.
     Inside(NodeId, Vec<Spanned<InsideRange>>),
     /// A function call such as `foo(a, b, c)`.
-    FunctionCall(NodeId, Vec<CallArg>),
+    FunctionCall(&'a ast::SubroutineDecl<'a>, &'a [ast::CallArg<'a>]),
     /// An assignment.
     Assign {
         op: ast::AssignOp,
         lhs: &'a ast::Expr<'a>,
         rhs: &'a ast::Expr<'a>,
     },
+    /// An expression in the AST that requires no representational change.
+    Ast(&'a ast::Expr<'a>),
 }
 
 /// The different unary operators.
@@ -942,14 +949,14 @@ impl HasDesc for Proc {
 
 /// A variable declaration.
 #[derive(Debug, PartialEq, Eq)]
-pub struct Stmt {
+pub struct Stmt<'a> {
     pub id: NodeId,
     pub label: Option<Spanned<Name>>,
     pub span: Span,
-    pub kind: StmtKind,
+    pub kind: StmtKind<'a>,
 }
 
-impl HasSpan for Stmt {
+impl HasSpan for Stmt<'_> {
     fn span(&self) -> Span {
         self.span
     }
@@ -959,7 +966,7 @@ impl HasSpan for Stmt {
     }
 }
 
-impl HasDesc for Stmt {
+impl HasDesc for Stmt<'_> {
     fn desc(&self) -> &'static str {
         #[allow(unreachable_patterns)]
         match self.kind {
@@ -973,7 +980,7 @@ impl HasDesc for Stmt {
 
 /// The different forms a statement can take.
 #[derive(Debug, Clone, PartialEq, Eq)]
-pub enum StmtKind {
+pub enum StmtKind<'a> {
     /// A null statement.
     Null,
     /// A sequential block.
@@ -1017,6 +1024,8 @@ pub enum StmtKind {
         default: Option<NodeId>,
         kind: ast::CaseKind,
     },
+    /// A statement in the AST that requires no representational change.
+    Ast(&'a ast::Stmt<'a>),
 }
 
 /// The different forms an assignment can take.
@@ -1291,52 +1300,3 @@ pub enum InsideRange {
     Single(NodeId),
     Range(NodeId, NodeId),
 }
-
-/// An argument to a function or method call.
-#[derive(Debug, Clone, Copy, PartialEq, Eq)]
-pub struct CallArg {
-    /// Span of the entire argument.
-    pub span: Span,
-    /// Optional argument name.
-    pub name: Option<Spanned<Name>>,
-    /// Value assigned to this argument.
-    pub expr: Option<NodeId>,
-}
-
-/// A subroutine declaration.
-#[derive(Debug, Clone, Copy, PartialEq, Eq)]
-pub struct Subroutine {
-    pub id: NodeId,
-    pub name: Spanned<Name>,
-    pub span: Span,
-    /// Whether this is a task or function.
-    pub kind: ast::SubroutineKind,
-    /// Optional return type in case of a function.
-    pub retty: Option<NodeId>,
-}
-
-impl HasSpan for Subroutine {
-    fn span(&self) -> Span {
-        self.span
-    }
-
-    fn human_span(&self) -> Span {
-        self.name.span
-    }
-}
-
-impl HasDesc for Subroutine {
-    fn desc(&self) -> &'static str {
-        match self.kind {
-            ast::SubroutineKind::Func => "function",
-            ast::SubroutineKind::Task => "task",
-        }
-    }
-
-    fn desc_full(&self) -> String {
-        match self.kind {
-            ast::SubroutineKind::Func => format!("function `{}`", self.name),
-            ast::SubroutineKind::Task => format!("task `{}`", self.name),
-        }
-    }
-}
