@@ -4,6 +4,7 @@
 
 use crate::crate_prelude::*;
 use crate::{
+    call_mapping::CallArgSource,
     hir::HirNode,
     mir::rvalue::*,
     syntax::ast::BasicNode,
@@ -12,7 +13,7 @@ use crate::{
     value::{self, ValueData, ValueKind},
     ParamEnv,
 };
-use num::{BigInt, One, Signed, ToPrimitive, Zero};
+use num::{BigInt, BigRational, One, Signed, ToPrimitive, Zero};
 use std::{cmp::max, collections::HashMap};
 
 /// An internal builder for rvalue lowering.
@@ -303,6 +304,7 @@ fn lower_expr_inner<'gcx>(
                     let k = builder.cx.constant_value_of(binding, env);
                     Ok(builder.build(ty, RvalueKind::Const(k)))
                 }
+                HirNode::SubroutinePort(port) => Ok(builder.build(ty, RvalueKind::Arg(port.id))),
                 x => {
                     builder.cx.emit(
                         DiagBuilder2::error(format!(
@@ -503,16 +505,15 @@ fn lower_expr_inner<'gcx>(
             Ok(check)
         }
 
-        hir::ExprKind::FunctionCall(..) => {
-            bug_span!(
-                span,
-                cx,
-                "lowering of {} to mir not yet supported",
-                hir.desc_full()
-            );
+        hir::ExprKind::FunctionCall(target, args) => {
+            Ok(lower_function_call(&builder, ty, target, args))
         }
 
         hir::ExprKind::Assign { op, lhs, rhs } => Ok(lower_assign(&builder, ty, op, lhs, rhs)),
+
+        hir::ExprKind::Ast(expr) => {
+            bug_span!(span, cx, "unsupported raw AST expr {:#?}", expr)
+        }
     }
 }
 
@@ -857,6 +858,18 @@ fn unpack_simple_bit_vector<'a>(
     }
     if to.coalesces_to_llhd_scalar() {
         builder.build(to, RvalueKind::Transmute(value))
+    } else if to.get_packed().map(|ty| ty.is_time()).unwrap_or(false) {
+        builder.cx.emit(
+            DiagBuilder2::warning(format!(
+                "assuming `1ns` timescale for conversion from `{}` to `time`",
+                value.ty
+            ))
+            .span(value.span),
+        );
+        builder.build(
+            to,
+            RvalueKind::ApplyTimescale(value, BigRational::new(1.into(), 1_000_000_000.into())),
+        )
     } else if let Some(dim) = to.outermost_dim() {
         unpack_array(builder, value, to, dim)
     } else if let Some(strukt) = to.get_struct() {
@@ -1665,3 +1678,66 @@ fn lower_assign<'a>(
         },
     )
 }
+
+/// Lower a function call.
+fn lower_function_call<'a>(
+    builder: &Builder<'_, impl Context<'a>>,
+    retty: &'a UnpackedType<'a>,
+    target: &'a ast::SubroutineDecl<'a>,
+    call_args: &'a [ast::CallArg<'a>],
+) -> &'a Rvalue<'a> {
+    trace!("Lowering call `{}`", builder.span.extract());
+    let cx = builder.cx;
+    // TODO(fschuiki): This should rather be some `call_details` query.
+    let decl_args = cx.canonicalize_func_args(Ref(target));
+    let mapping = cx.call_mapping(Ref(decl_args), Ref(call_args), builder.span);
+    if mapping.is_error() {
+        return builder.error();
+    }
+
+    // Decide how we are going to pass values in and out of the called function.
+    // This must match how the function code generator will eventually emit the
+    // function in
+    let mut args = vec![];
+    for arg in &mapping.args {
+        trace!("Lowering call argument `{}`", arg.span.extract());
+        match arg.decl.dir {
+            ast::SubroutinePortDir::Input => {
+                let rv = cx.mir_rvalue(arg.expr.id(), arg.env);
+                args.push(CallArg::Input(rv));
+            }
+            ast::SubroutinePortDir::Output => {
+                // Output arguments with a default value have no place to copy
+                // the data out to after the call.
+                let ty = cx.type_of_func_arg(Ref(arg.decl), arg.env);
+                let lv = match arg.call {
+                    CallArgSource::Call(..) => Some(cx.mir_lvalue(arg.expr.id(), arg.env)),
+                    CallArgSource::Default(..) => None,
+                };
+                args.push(CallArg::Output(ty, lv));
+            }
+            ast::SubroutinePortDir::Inout => {
+                // Output arguments with a default value have no place to
+                // copy the data out to after the call.
+                let lv = match arg.call {
+                    CallArgSource::Call(..) => Some(cx.mir_lvalue(arg.expr.id(), arg.env)),
+                    CallArgSource::Default(..) => None,
+                };
+                let rv = cx.mir_rvalue(arg.expr.id(), arg.env);
+                args.push(CallArg::Inout(rv, lv));
+            }
+            ast::SubroutinePortDir::Ref | ast::SubroutinePortDir::ConstRef => {
+                let lv = cx.mir_lvalue(arg.expr.id(), arg.env);
+                args.push(CallArg::Ref(lv));
+            }
+        }
+    }
+
+    trace!("Call args:");
+    for arg in &args {
+        trace!("  - {:?}", arg);
+    }
+
+    // Package things up into a call.
+    builder.build(retty, RvalueKind::Call { target, args })
+}
