@@ -15,6 +15,7 @@ use crate::{
     ty::{Domain, Sign, UnpackedType},
     ParamEnv,
 };
+use num::BigRational;
 use std::collections::HashMap;
 use std::fmt::Write;
 
@@ -90,6 +91,12 @@ impl<'a> Print for Rvalue<'a> {
                 write!(inner, "CastSign({}, {})", sign, ctx.print(outer, arg))?
             }
             RvalueKind::CastToBool(arg) => write!(inner, "CastToBool({})", ctx.print(outer, arg))?,
+            RvalueKind::ApplyTimescale(arg, ref scale) => write!(
+                inner,
+                "ApplyTimescale({}, {})",
+                ctx.print(outer, arg),
+                scale
+            )?,
             RvalueKind::Truncate(size, arg) => {
                 write!(inner, "Truncate({}, {})", size, ctx.print(outer, arg))?
             }
@@ -171,6 +178,7 @@ impl<'a> Print for Rvalue<'a> {
             }
             RvalueKind::Var(arg) => write!(inner, "Var({:?})", arg)?,
             RvalueKind::Port(arg) => write!(inner, "Port({:?})", arg)?,
+            RvalueKind::Arg(arg) => write!(inner, "Arg({:?})", arg)?,
             RvalueKind::Intf(arg) => write!(inner, "Intf({:?})", arg)?,
             RvalueKind::IntfSignal(arg, sig) => {
                 write!(inner, "IntfSignal({}, {:?})", ctx.print(outer, arg), sig)?
@@ -249,6 +257,14 @@ impl<'a> Print for Rvalue<'a> {
                 op,
                 ctx.print(outer, rhs)
             )?,
+            RvalueKind::Call { target, ref args } => {
+                write!(
+                    inner,
+                    "Call {} ({})",
+                    target.prototype.name,
+                    ctx.print_comma_separated(outer, args),
+                )?;
+            }
             RvalueKind::Error => write!(inner, "<error>")?,
         }
         write!(inner, " : {}", self.ty)?;
@@ -357,6 +373,8 @@ pub enum RvalueKind<'a> {
     Var(NodeId),
     /// A reference to a port declaration.
     Port(NodeId),
+    /// A reference to a subroutine argument.
+    Arg(NodeId),
     /// A reference to an interface.
     Intf(NodeId),
     /// A reference to a locally instantiated interface signal.
@@ -407,6 +425,16 @@ pub enum RvalueKind<'a> {
         lhs: &'a Rvalue<'a>,
         rhs: &'a Rvalue<'a>,
     },
+    /// Convert an integer to a time value by applying the currently active timescale.
+    ApplyTimescale(&'a Rvalue<'a>, BigRational),
+    /// A function or task call.
+    Call {
+        /// The called function.
+        #[dont_visit]
+        target: &'a ast::SubroutineDecl<'a>,
+        /// The call arguments.
+        args: Vec<CallArg<'a>>,
+    },
     /// An error occurred during lowering.
     Error,
 }
@@ -433,7 +461,8 @@ impl<'a> RvalueKind<'a> {
             | RvalueKind::Repeat(_, value)
             | RvalueKind::Member { value, .. }
             | RvalueKind::PackString(value)
-            | RvalueKind::UnpackString(value) => value.is_const(),
+            | RvalueKind::UnpackString(value)
+            | RvalueKind::ApplyTimescale(value, _) => value.is_const(),
             RvalueKind::ConstructArray(values) => values.values().all(|v| v.is_const()),
             RvalueKind::ConstructStruct(values) => values.iter().all(|v| v.is_const()),
             RvalueKind::Const(_) => true,
@@ -447,6 +476,7 @@ impl<'a> RvalueKind<'a> {
             RvalueKind::Concat(values) => values.iter().all(|v| v.is_const()),
             RvalueKind::Var(_) => false,
             RvalueKind::Port(_) => false,
+            RvalueKind::Arg(_) => false,
             RvalueKind::Intf(_) => false,
             RvalueKind::IntfSignal(..) => false,
             RvalueKind::Index { .. } => false, // TODO(fschuiki): reactivate once impl
@@ -458,6 +488,9 @@ impl<'a> RvalueKind<'a> {
             } => cond.is_const() && true_value.is_const() && false_value.is_const(),
             RvalueKind::Shift { value, amount, .. } => value.is_const() && amount.is_const(),
             RvalueKind::Assignment { .. } => false,
+            // TODO(fschuiki): This is wrong; function calls *may* be constant
+            // under certain circumstances.
+            RvalueKind::Call { .. } => false,
             RvalueKind::Error => true,
         }
     }
@@ -532,3 +565,76 @@ pub enum ShiftOp {
     Left,
     Right,
 }
+
+/// A call input argument.
+#[moore_derive::visit_without_foreach]
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub enum CallInput<'a> {
+    /// The argument is passed by value.
+    ByValue(&'a Rvalue<'a>),
+    /// The argument is passed by reference.
+    ByRef(&'a Lvalue<'a>),
+}
+
+impl<'a> Print for CallInput<'a> {
+    fn print_context(
+        &self,
+        outer: &mut impl Write,
+        inner: &mut impl Write,
+        ctx: &mut Context,
+    ) -> std::fmt::Result {
+        match self {
+            Self::ByValue(x) => x.print_context(outer, inner, ctx),
+            Self::ByRef(x) => x.print_context(outer, inner, ctx),
+        }
+    }
+}
+
+/// A call argument.
+#[moore_derive::visit_without_foreach]
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub enum CallArg<'a> {
+    /// An input argument. Passed to the function by value.
+    Input(&'a Rvalue<'a>),
+    /// An output argument. Passed as a temporary variable to the function. The
+    /// after the function returns, a post-assignment in the `RvalueKind::Call`
+    /// is responsible for transfering the value from the temporary variable to
+    /// the actual `Lvalue` listed here, if any. This ensures signals passed as
+    /// output arguments get a proper blocking assignment.
+    Output(&'a UnpackedType<'a>, Option<&'a Lvalue<'a>>),
+    /// An inout argument. Same as `CallArg::Output`, but the variable is
+    /// initialized to the provided value.
+    Inout(&'a Rvalue<'a>, Option<&'a Lvalue<'a>>),
+    /// A by-ref argument. Passed into the function as-is.
+    Ref(&'a Lvalue<'a>),
+}
+
+impl<'a> Print for CallArg<'a> {
+    fn print_context(
+        &self,
+        outer: &mut impl Write,
+        inner: &mut impl Write,
+        ctx: &mut Context,
+    ) -> std::fmt::Result {
+        match self {
+            Self::Input(x) => {
+                write!(inner, "in:")?;
+                x.print_context(outer, inner, ctx)
+            }
+            Self::Output(_, x) => {
+                write!(inner, "out:")?;
+                x.print_context(outer, inner, ctx)
+            }
+            Self::Inout(x, y) => {
+                write!(inner, "in:")?;
+                x.print_context(outer, inner, ctx)?;
+                write!(inner, "/")?;
+                y.print_context(outer, inner, ctx)
+            }
+            Self::Ref(x) => {
+                write!(inner, "ref:")?;
+                x.print_context(outer, inner, ctx)
+            }
+        }
+    }
+}
